package ru.dbotthepony.mc.otm.block.entity

import com.google.common.collect.ImmutableSet
import it.unimi.dsi.fastutil.longs.Long2ObjectFunction
import it.unimi.dsi.fastutil.longs.Long2ObjectOpenHashMap
import it.unimi.dsi.fastutil.objects.Object2ObjectArrayMap
import it.unimi.dsi.fastutil.objects.Reference2ObjectArrayMap
import net.minecraft.client.multiplayer.ClientLevel
import net.minecraft.core.BlockPos
import net.minecraft.core.Direction
import net.minecraft.core.SectionPos
import net.minecraft.nbt.ByteTag
import net.minecraft.nbt.CompoundTag
import net.minecraft.nbt.DoubleTag
import net.minecraft.nbt.FloatTag
import net.minecraft.nbt.StringTag
import net.minecraft.nbt.Tag
import net.minecraft.resources.ResourceLocation
import net.minecraft.server.level.ServerLevel
import net.minecraft.server.level.ServerPlayer
import net.minecraft.world.level.ChunkPos
import net.minecraft.world.level.Level
import net.minecraft.world.level.block.entity.BlockEntity
import net.minecraft.world.level.block.entity.BlockEntityType
import net.minecraft.world.level.block.entity.RandomizableContainerBlockEntity
import net.minecraft.world.level.block.state.BlockState
import net.minecraftforge.common.capabilities.Capability
import net.minecraftforge.common.capabilities.ForgeCapabilities
import net.minecraftforge.common.util.INBTSerializable
import net.minecraftforge.common.util.LazyOptional
import net.minecraftforge.event.TickEvent
import net.minecraftforge.event.TickEvent.LevelTickEvent
import net.minecraftforge.event.entity.player.PlayerEvent
import net.minecraftforge.event.level.ChunkWatchEvent
import net.minecraftforge.event.level.LevelEvent
import net.minecraftforge.event.server.ServerStoppingEvent
import net.minecraftforge.items.IItemHandler
import ru.dbotthepony.mc.otm.block.RotatableMatteryBlock
import ru.dbotthepony.mc.otm.capability.FlowDirection
import ru.dbotthepony.mc.otm.capability.MatteryCapability
import ru.dbotthepony.mc.otm.capability.energy.IMatteryEnergyStorage
import ru.dbotthepony.mc.otm.capability.isMekanismLoaded
import ru.dbotthepony.mc.otm.compat.mekanism.Mattery2MekanismEnergyWrapper
import ru.dbotthepony.mc.otm.core.forValidRefs
import ru.dbotthepony.mc.otm.core.forValidRefsBreak
import ru.dbotthepony.mc.otm.core.get
import ru.dbotthepony.mc.otm.core.math.BlockRotation
import ru.dbotthepony.mc.otm.core.math.RelativeSide
import ru.dbotthepony.mc.otm.core.nbt.set
import ru.dbotthepony.mc.otm.core.util.EnumValueCodec
import ru.dbotthepony.mc.otm.network.BlockEntitySyncPacket
import ru.dbotthepony.mc.otm.network.FieldSynchronizer
import ru.dbotthepony.mc.otm.network.WorldNetworkChannel
import ru.dbotthepony.mc.otm.oncePre
import ru.dbotthepony.mc.otm.onceServer
import java.lang.ref.WeakReference
import java.util.Collections
import java.util.EnumMap
import java.util.WeakHashMap
import kotlin.properties.ReadWriteProperty
import kotlin.reflect.KMutableProperty0
import kotlin.reflect.KProperty
import kotlin.reflect.KProperty0

/**
 * Absolute barebone block entity class in Overdrive that Matters, providing bare minimum functionality
 */
abstract class MatteryBlockEntity(p_155228_: BlockEntityType<*>, p_155229_: BlockPos, p_155230_: BlockState) : BlockEntity(p_155228_, p_155229_, p_155230_) {
	private var isSynchronizing = false

	/**
	 * "shortcut" for getting [BlockRotation]
	 *
	 * if block has no rotation, returns [BlockRotation.NORTH]
	 */
	open val blockRotation: BlockRotation get() {
		return blockState[(blockState.block as? RotatableMatteryBlock ?: return BlockRotation.NORTH).rotationProperty]
	}

	private val _sides = EnumMap<RelativeSide, Side>(RelativeSide::class.java)
	val sides: Map<RelativeSide, Side> = Collections.unmodifiableMap(_sides)

	private data class SidelessCap<T : Any>(val cap: T, var optional: LazyOptional<T>)
	private val sidelessCaps = Reference2ObjectArrayMap<Capability<*>, SidelessCap<*>>()

	/**
	 * exposes capability when no side is specified
	 */
	protected fun <T : Any> exposeSideless(capability: Capability<T>, value: T) {
		check(!sidelessCaps.containsKey(capability)) { "Already has globally exposed $capability!" }
		sidelessCaps[capability] = SidelessCap(value, LazyOptional.of { value })
		setChanged()
	}

	protected fun <T : Any> exposeAllSidesExcept(side: RelativeSide, capability: Capability<T>, value: T) {
		for ((k, v) in _sides) {
			if (k != side) {
				v.Cap(capability, value)
			}
		}
	}

	protected fun isSidelessExposed(capability: Capability<*>): Boolean {
		return sidelessCaps.containsKey(capability)
	}

	protected fun removeSideless(capability: Capability<*>): Boolean {
		if (sidelessCaps.remove(capability) != null) {
			setChanged()
			return true
		}

		return false
	}

	/**
	 * Exposes capability unconditionally, on all sides and sideless
	 */
	protected fun <T : Any> exposeGlobally(capability: Capability<T>, value: T) {
		exposeSideless(capability, value)

		for (side in _sides.values)
			side.Cap(capability, value)
	}

	protected fun exposeEnergyGlobally(value: IMatteryEnergyStorage) {
		exposeGlobally(ForgeCapabilities.ENERGY, value)
		exposeGlobally(MatteryCapability.ENERGY, value)

		if (isMekanismLoaded) {
			exposeGlobally(MatteryCapability.MEKANISM_ENERGY, Mattery2MekanismEnergyWrapper(value))
		}
	}

	protected fun exposeEnergy(side: RelativeSide, value: IMatteryEnergyStorage) {
		_sides[side]!!.Cap(ForgeCapabilities.ENERGY, value)
		_sides[side]!!.Cap(MatteryCapability.ENERGY, value)

		if (isMekanismLoaded) {
			_sides[side]!!.Cap(MatteryCapability.MEKANISM_ENERGY, Mattery2MekanismEnergyWrapper(value))
		}
	}

	protected fun exposeEnergySideless( value: IMatteryEnergyStorage) {
		exposeSideless(ForgeCapabilities.ENERGY, value)
		exposeSideless(MatteryCapability.ENERGY, value)

		if (isMekanismLoaded) {
			exposeSideless(MatteryCapability.MEKANISM_ENERGY, Mattery2MekanismEnergyWrapper(value))
		}
	}

	protected fun exposeItemsGlobally(value: IItemHandler) {
		exposeGlobally(ForgeCapabilities.ITEM_HANDLER, value)
	}

	enum class SideMode(val filter: FlowDirection, val automation: FlowDirection) {
		DISABLED		(FlowDirection.NONE, 			FlowDirection.NONE),
		NONE			(FlowDirection.BI_DIRECTIONAL, 	FlowDirection.NONE),
		INPUT			(FlowDirection.INPUT, 			FlowDirection.NONE),
		OUTPUT			(FlowDirection.OUTPUT, 			FlowDirection.NONE),

		PULL			(FlowDirection.INPUT, 			FlowDirection.INPUT),
		PUSH			(FlowDirection.OUTPUT, 			FlowDirection.OUTPUT),
		PULL_ONLY		(FlowDirection.BI_DIRECTIONAL, 	FlowDirection.INPUT),
		PUSH_ONLY		(FlowDirection.BI_DIRECTIONAL, 	FlowDirection.OUTPUT),
		PULL_PUSH		(FlowDirection.BI_DIRECTIONAL, 	FlowDirection.BI_DIRECTIONAL);

		val isOpen = filter != FlowDirection.NONE

		companion object {
			@JvmField
			val BI_SET: ImmutableSet<SideMode> = ImmutableSet.of(NONE, INPUT, OUTPUT, PULL, PUSH, PULL_ONLY, PUSH_ONLY, PULL_PUSH, DISABLED)

			@JvmField
			val INPUT_SET: ImmutableSet<SideMode> = ImmutableSet.of(INPUT, PULL, DISABLED)

			@JvmField
			val OUTPUT_SET: ImmutableSet<SideMode> = ImmutableSet.of(OUTPUT, PUSH, DISABLED)

			@JvmField
			val CODEC = EnumValueCodec(SideMode::class.java)
		}
	}

	fun interface SideModeListener {
		fun sideModeChanges(config: Side.ModeState, new: SideMode, old: SideMode)
	}

	inner class Side(val side: RelativeSide) : INBTSerializable<CompoundTag> {
		init {
			check(!_sides.containsKey(side)) { "dafuq are you trying to do" }
			_sides[side] = this
		}

		private val caps = Reference2ObjectArrayMap<Capability<*>, Cap<*>>()
		private val modeStates = Object2ObjectArrayMap<String, ModeState>()
		private val data = Object2ObjectArrayMap<ResourceLocation, INBTSerializable<Tag>>()

		val isEmpty get() = caps.isEmpty() && modeStates.isEmpty()

		operator fun <T : Any> get(capability: Capability<T>): Cap<T>? {
			return caps[capability] as Cap<T>?
		}

		operator fun get(index: String): ModeState? {
			return modeStates[index]
		}

		fun addData(index: ResourceLocation, data: INBTSerializable<*>) {
			this.data[index] = data as INBTSerializable<Tag>
		}

		fun getData(index: ResourceLocation): INBTSerializable<*>? {
			return data[index]
		}

		fun invalidate() {
			for (cap in caps.values)
				cap.invalidate()
		}

		fun revive() {
			for (cap in caps.values)
				cap.revive()
		}

		override fun serializeNBT(): CompoundTag {
			return CompoundTag().also {
				val modelist = CompoundTag()

				for (cap in modeStates.values) {
					modelist[cap.name] = cap.serializeNBT()
				}

				it["Modes"] = modelist
			}
		}

		override fun deserializeNBT(nbt: CompoundTag) {
			if (nbt.contains("Modes")) {
				val caplist = nbt.getCompound("Modes")

				for (state in modeStates.values) {
					if (state.name in caplist) {
						state.deserializeNBT(caplist.getString(state.name))
					}
				}
			}
		}

		inner class ModeState(val name: String, val possibleValues: ImmutableSet<SideMode> = SideMode.BI_SET) : INBTSerializable<StringTag> {
			init {
				check(!modeStates.containsKey(name)) { "Already has config with name $name on side $side" }
				modeStates[name] = this
			}

			val side get() = this@Side.side

			var mode by synchronizer.Field(possibleValues.first(), SideMode.CODEC, setter = { value, access, setByRemote ->
				require(value in possibleValues) { "Value $value is not allowed (allowed values: $possibleValues)" }
				val old = access.read()

				if (value != old) {
					access.write(value)

					listeners.forValidRefs {
						it.sideModeChanges(this, value, old)
					}
				}
			}, name = "SideConfig/$side/$name")

			fun isAllowed(value: SideMode): Boolean = value in possibleValues

			private val listeners = ArrayList<WeakReference<SideModeListener>>()

			fun addListener(listener: SideModeListener) {
				var hit = false

				listeners.forValidRefsBreak {
					if (it === listener) {
						hit = true
						return@forValidRefsBreak true
					}

					return@forValidRefsBreak false
				}

				if (!hit) {
					listeners.add(WeakReference(listener))
				}
			}

			override fun serializeNBT(): StringTag {
				return StringTag.valueOf(mode.name)
			}

			override fun deserializeNBT(nbt: StringTag) {
				mode = SideMode.valueOf(nbt.asString)
			}

			fun deserializeNBT(nbt: String) {
				mode = SideMode.valueOf(nbt)
			}
		}

		inner class Cap<T : Any>(val type: Capability<in T>, val capability: T) {
			init {
				check(!caps.containsKey(type)) { "Already has capability $type on side $side" }
				caps[type] = this
			}

			private var isExposed = true
			private var isValid = true
			private var isRemoved = false

			var optional: LazyOptional<T> by object : ReadWriteProperty<Any?, LazyOptional<T>> {
				private var value: LazyOptional<T>? = null

				override fun getValue(thisRef: Any?, property: KProperty<*>): LazyOptional<T> {
					if (value == null) {
						value = LazyOptional.of { capability }
					}

					return value!!
				}

				override fun setValue(thisRef: Any?, property: KProperty<*>, value: LazyOptional<T>) {
					this.value = value
				}
			}
				private set

			fun remove() {
				if (!isRemoved) {
					isRemoved = true
					val removed = caps.remove(type)
					check(removed == this) { "$removed != $this" }
					optional.invalidate()
				}
			}

			fun close() {
				if (!isRemoved && isExposed) {
					isExposed = false
					optional.invalidate()
					setChanged()
				}
			}

			fun expose() {
				if (!isRemoved && !isExposed) {
					isExposed = true

					if (isValid) {
						optional = LazyOptional.of { capability }
						setChanged()
					}
				}
			}

			fun invalidate() {
				if (!isRemoved && isValid) {
					isValid = false
					optional.invalidate()
					setChanged()
				}
			}

			fun revive() {
				if (!isRemoved && !isValid) {
					isValid = true

					if (isExposed) {
						optional = LazyOptional.of { capability }
						setChanged()
					}
				}
			}
		}
	}

	val front = Side(RelativeSide.FRONT)
	val back = Side(RelativeSide.BACK)
	val left = Side(RelativeSide.LEFT)
	val right = Side(RelativeSide.RIGHT)
	val top = Side(RelativeSide.TOP)
	val bottom = Side(RelativeSide.BOTTOM)

	override fun <T : Any> getCapability(cap: Capability<T>, side: Direction?): LazyOptional<T> {
		if (side != null) {
			val get = _sides[blockRotation.dir2Side(side)]!![cap]

			if (get != null)
				return get.optional

			return super.getCapability(cap, side)
		}

		return sidelessCaps[cap]?.optional?.cast() ?: super.getCapability(cap, side)
	}

	override fun invalidateCaps() {
		super.invalidateCaps()

		for (side in sides.values)
			side.invalidate()
	}

	override fun reviveCaps() {
		super.reviveCaps()

		for (side in sides.values)
			side.revive()
	}

	private data class Savetable0<T : Tag, V : Any>(val type: Class<T>, val property: KProperty0<V>, val name: String, val serialize: (V) -> T, val deserialize: (V, T) -> Unit)
	private data class Savetable1<T : Tag, V : Any>(val type: Class<T>, val property: KMutableProperty0<V>, val name: String, val serialize: (V) -> T, val deserialize: (T) -> V)

	private val savetables0 = ArrayList<Savetable0<*, *>>()
	private val savetables1 = ArrayList<Savetable1<*, *>>()

	/**
	 * Save and load object state
	 */
	protected fun <T : Tag, V : Any> savetable(type: Class<T>, property: KProperty0<V>, name: String = property.name, serialize: (V) -> T, deserialize: (V, T) -> Unit) {
		check(!savetables0.any { it.name == name }) { "Already has save field with name $name" }
		check(!savetables1.any { it.name == name }) { "Already has save field with name $name" }
		savetables0.add(Savetable0(type, property, name, serialize, deserialize))
	}

	/**
	 * Save and load value
	 */
	protected fun <T : Tag, V : Any> savetable(type: Class<T>, property: KMutableProperty0<V>, name: String = property.name, serialize: (V) -> T, deserialize: (T) -> V) {
		check(!savetables0.any { it.name == name }) { "Already has save field with name $name" }
		check(!savetables1.any { it.name == name }) { "Already has save field with name $name" }
		savetables1.add(Savetable1(type, property, name, serialize, deserialize))
	}

	/**
	 * Save and load object state
	 */
	protected fun <T : Tag, V : INBTSerializable<T?>> savetable(type: Class<T>, property: KProperty0<V>, name: String = property.name) {
		savetable(type, property, name, { it.serializeNBT()!! }, { self, it -> self.deserializeNBT(it) })
	}

	/**
	 * Save and load object state
	 */
	protected inline fun <reified T : Tag, V : INBTSerializable<T?>> savetable(property: KProperty0<V>, name: String = property.name) {
		savetable(T::class.java, property, name)
	}

	/**
	 * Save and load object state
	 */
	protected inline fun <reified T : Tag, V : Any> savetable(property: KProperty0<V>, name: String = property.name, noinline serialize: (V) -> T, noinline deserialize: (V, T) -> Unit) {
		savetable(T::class.java, property, name, serialize, deserialize)
	}

	/**
	 * Save and load value
	 */
	protected inline fun <reified T : Tag, V : Any> savetable(property: KMutableProperty0<V>, name: String = property.name, noinline serialize: (V) -> T, noinline deserialize: (T) -> V) {
		savetable(T::class.java, property, name, serialize, deserialize)
	}

	protected fun savetableFloat(property: KMutableProperty0<Float>, name: String = property.name) {
		savetable(FloatTag::class.java, property, name, { FloatTag.valueOf(it) }, { tag -> tag.asFloat })
	}

	protected fun savetableDouble(property: KMutableProperty0<Double>, name: String = property.name) {
		savetable(DoubleTag::class.java, property, name, { DoubleTag.valueOf(it) }, { tag -> tag.asDouble })
	}

	protected fun savetableBoolean(property: KMutableProperty0<Boolean>, name: String = property.name) {
		savetable(ByteTag::class.java, property, name, { if (it) ByteTag.ONE else ByteTag.ZERO }, { tag -> tag.asInt > 0 })
	}

	override fun saveAdditional(nbt: CompoundTag) {
		super.saveAdditional(nbt)

		if (_sides.values.any { !it.isEmpty }) {
			nbt["Sides"] = CompoundTag().also {
				for (side in _sides.values) {
					if (!side.isEmpty) {
						it[side.side.name] = side.serializeNBT()
					}
				}
			}
		}

		for (data in savetables0) {
			nbt[data.name] = (data.serialize as (Any) -> Tag).invoke(data.property.get())
		}

		for (data in savetables1) {
			nbt[data.name] = (data.serialize as (Any) -> Tag).invoke(data.property.get())
		}
	}

	override fun load(nbt: CompoundTag) {
		super.load(nbt)

		if (nbt.contains("Sides")) {
			val tag = nbt.getCompound("Sides")

			for (side in _sides.values) {
				if (side.side.name in tag) {
					side.deserializeNBT(tag.getCompound(side.side.name))
				}
			}
		}

		for (data in savetables0) {
			val value = nbt[data.name]

			if (value != null && data.type.isAssignableFrom(value.javaClass)) {
				(data.deserialize as (Any, Tag) -> Unit).invoke(data.property.get(), value)
			} else if (value != null) {
				throw ClassCastException("Expected ${data.type.canonicalName}, got ${value.javaClass.canonicalName}")
			}
		}

		for (data in savetables1) {
			val value = nbt[data.name]

			if (value != null && data.type.isAssignableFrom(value.javaClass)) {
				(data.property as KMutableProperty0<Any>).set((data.deserialize as (Tag) -> Any).invoke(value))
			} else if (value != null) {
				throw ClassCastException("Expected ${data.type.canonicalName}, got ${value.javaClass.canonicalName}")
			}
		}
	}

	protected fun tickOnce(func: Runnable) {
		level?.oncePre { if (!isRemoved) func.run() }
	}

	protected fun tickOnceServer(func: Runnable) {
		(level as? ServerLevel)?.oncePre { if (!isRemoved) func.run() }
	}

	protected fun tickOnceClient(func: Runnable) {
		(level as? ClientLevel)?.oncePre { if (!isRemoved) func.run() }
	}

	protected fun tickOnce(func: (Level) -> Unit) {
		val level = level
		level?.oncePre { if (!isRemoved) func.invoke(level) }
	}

	protected fun tickOnceServer(func: (ServerLevel) -> Unit) {
		val level = level as? ServerLevel ?: return
		level.oncePre { if (!isRemoved) func.invoke(level) }
	}

	protected fun tickOnceClient(func: (ClientLevel) -> Unit) {
		val level = level as? ClientLevel ?: return
		level.oncePre { if (!isRemoved) func.invoke(level) }
	}

	// Just to mark chunk unsaved
	open fun setChangedLight() {
		val level = level
		if (level is ServerLevel) {
			level.chunkSource.getChunkNow(
				SectionPos.blockToSectionCoord(blockPos.x),
				SectionPos.blockToSectionCoord(blockPos.z))?.isUnsaved = true
		}
	}

	val synchronizer = FieldSynchronizer {
		if (isSynchronizing)
			return@FieldSynchronizer

		if (!isRemoved && level?.isClientSide == false && (_subCache == null || (_subCache ?: throw ConcurrentModificationException()).players.isNotEmpty())) {
			onceServer {
				synchronizeToPlayers(true)
			}
		} else {
			markSynchronizerClean()
		}
	}

	private fun markSynchronizerClean() {
		synchronizer.markClean()
	}

	init {
		synchronizer.defaultEndpoint.markUnused()
	}

	override fun setLevel(p_155231_: Level) {
		super.setLevel(p_155231_)
		unsubscribe()
		_subCache = null

		if (!p_155231_.isClientSide) {
			subscribe()
		}
	}

	override fun setRemoved() {
		super.setRemoved()
		unsubscribe()
	}

	override fun clearRemoved() {
		super.clearRemoved()

		if (level?.isClientSide == false) {
			subscribe()
		}
	}

	private var _subCache: ChunkSubscribers? = null

	private fun unsubscribe() {
		val subCache = _subCache ?: return

		if (subCache.unsubscribe(this)) {
			val level = subCache.level.get()

			if (level != null) {
				playerMap.get(level)?.remove(subCache.chunkPos)
			}
		}

		lastSubscriptionChangeset = -1
	}

	private fun subscribe(): ChunkSubscribers {
		val level = level
		check(level is ServerLevel) { "Invalid realm" }
		unsubscribe()

		val subs = playerMap
			.computeIfAbsent(level) { Long2ObjectOpenHashMap() }
			.computeIfAbsent(ChunkPos(blockPos).toLong(), Long2ObjectFunction { ChunkSubscribers(level = WeakReference(level), chunkPos = ChunkPos(blockPos).toLong()) })

		subs.subscribe(this)
		_subCache = subs
		return subs
	}

	private val subscription get() = _subCache ?: subscribe()
	private var lastSubscriptionChangeset = -1

	fun synchronizeToPlayers() {
		synchronizeToPlayers(false)
	}

	protected fun synchronizeToPlayers(calledBySynchronizer: Boolean) {
		isSynchronizing = true

		try {
			if (synchronizer.isEmpty || isRemoved) {
				if (calledBySynchronizer) synchronizer.markClean()
				return
			}

			check(level is ServerLevel) { "Invalid realm or Level is null" }

			synchronizer.observe()

			val subscription = subscription

			if (subscription.players.isEmpty() || lastSubscriptionChangeset == subscription.changeset && !synchronizer.hasChanges) {
				if (calledBySynchronizer) synchronizer.markClean()
				return
			}

			lastSubscriptionChangeset = subscription.changeset

			for (player in subscription.players) {
				val payload = synchronizer.computeEndpointFor(player).collectNetworkPayload()

				if (payload != null) {
					WorldNetworkChannel.send(player, BlockEntitySyncPacket(blockPos, payload.array, payload.length))
				}
			}

			synchronizer.markClean()
		} finally {
			isSynchronizing = false
		}
	}

	private data class ChunkSubscribers(
		val blockEntities: ArrayList<WeakReference<MatteryBlockEntity>> = ArrayList(0),
		val players: ArrayList<ServerPlayer> = ArrayList(1),
		val observingBlockEntities: ArrayList<WeakReference<MatteryBlockEntity>> = ArrayList(0),
		val level: WeakReference<ServerLevel>,
		val chunkPos: Long,
		var changeset: Int = 0
	) {
		fun cleanUpBlocks() {
			val listIterator = blockEntities.listIterator()

			for (block in listIterator) {
				if (block.get() == null) {
					listIterator.remove()
				}
			}
		}

		val hasObservers: Boolean get() {
			val listIterator = blockEntities.listIterator()

			for (value in listIterator) {
				val ref = value.get()

				if (ref == null) {
					listIterator.remove()
				} else if (ref.synchronizer.hasObservers) {
					return true
				}
			}

			return false
		}

		fun subscribe(blockEntity: MatteryBlockEntity) {
			val iterator = blockEntities.listIterator()

			for (value in iterator) {
				val ref = value.get()

				if (ref === blockEntity) {
					return
				} else if (ref == null) {
					iterator.remove()
				}
			}

			val bref = WeakReference(blockEntity)
			blockEntities.add(bref)
			changeset++

			onceServer {
				blockEntity.synchronizeToPlayers()
			}

			if (blockEntity.synchronizer.hasObservers) {
				observingBlockEntities.add(bref)
				val listing = tickingMap.computeIfAbsent(level.get() ?: throw NullPointerException("Level got GCd!")) { ArrayList(1) }
				val tickerIterator = listing.listIterator()

				for (value in tickerIterator) {
					val ref = value.get()

					if (ref === this) {
						return
					} else if (ref == null) {
						tickerIterator.remove()
					}
				}

				listing.add(WeakReference(this))
			}
		}

		fun unsubscribe(blockEntity: MatteryBlockEntity): Boolean {
			val listIterator = blockEntities.listIterator()

			for (value in listIterator) {
				if (value.get() === blockEntity) {
					listIterator.remove()
					changeset++
					break
				}
			}

			if (blockEntity.synchronizer.hasObservers) {
				val tickerIterator = observingBlockEntities.listIterator()

				for (value in tickerIterator) {
					val ref = value.get()

					if (ref === blockEntity) {
						tickerIterator.remove()
						break
					} else if (ref == null) {
						tickerIterator.remove()
					}
				}
			}

			if (players.isNotEmpty()) {
				return false
			}

			cleanUpBlocks()
			return blockEntities.isEmpty()
		}

		val isEmpty: Boolean get() {
			if (players.isNotEmpty()) {
				return false
			}

			cleanUpBlocks()
			return blockEntities.isEmpty()
		}
	}

	/**
	 * Why track player-tracked chunks?
	 *
	 * because minecraft itself doesn't track them, well,
	 * in "section partitioning" way.
	 *
	 * just look at [net.minecraft.server.level.PlayerMap]
	 *
	 * the [net.minecraft.server.level.PlayerMap.getPlayers] straight ignores chunk position
	 * and just fetches all players
	 *
	 * even if they did not ignore that argument, you still have to fetch player *list* though
	 * [net.minecraft.server.level.ChunkMap.getPlayers], which is not something we want
	 */
	companion object {
		const val REDSTONE_CONTROL_KEY = "redstoneControl"
		const val INVENTORY_KEY = "container"
		const val MATTER_STORAGE_KEY = "matterStorage"
		const val ENERGY_KEY = "energyStorage"
		const val BATTERY_KEY = "batteryInventory"
		const val LOOT_TABLE_KEY = RandomizableContainerBlockEntity.LOOT_TABLE_TAG
		const val LOOT_TABLE_SEED_KEY = RandomizableContainerBlockEntity.LOOT_TABLE_SEED_TAG

		private val playerMap = WeakHashMap<ServerLevel, Long2ObjectOpenHashMap<ChunkSubscribers>>()
		private val tickingMap = WeakHashMap<ServerLevel, ArrayList<WeakReference<ChunkSubscribers>>>()

		fun onLevelUnload(event: LevelEvent.Unload) {
			val level = event.level as? ServerLevel ?: return
			playerMap.remove(level)
			tickingMap.remove(level)
		}

		fun onServerStopping(event: ServerStoppingEvent) {
			playerMap.clear()
			tickingMap.clear()
		}

		fun postLevelTick(event: LevelTickEvent) {
			if (event.phase == TickEvent.Phase.END) {
				val level = event.level as? ServerLevel ?: return
				val listing = tickingMap[level] ?: return
				val listIterator = listing.listIterator()
				var hitAnyObservers = false

				for (value in listIterator) {
					val ref = value.get()

					if (ref == null) {
						listIterator.remove()
					} else if (ref.players.isNotEmpty()) {
						var hitObservers = false

						ref.observingBlockEntities.forValidRefs {
							hitObservers = true
							hitAnyObservers = true
							it.synchronizeToPlayers()
						}

						if (!hitObservers) {
							listIterator.remove()
						}
					}
				}

				if (!hitAnyObservers) {
					tickingMap.remove(level)
				}
			}
		}

		fun onWatch(event: ChunkWatchEvent.Watch) {
			playerMap
				.computeIfAbsent(event.level) { Long2ObjectOpenHashMap() }
				.computeIfAbsent(event.pos.toLong(), Long2ObjectFunction { ChunkSubscribers(level = WeakReference(event.level), chunkPos = event.pos.toLong()) })
			.let {
				val (blocks, players) = it

				if (event.player !in players) {
					players.add(event.player)
					it.changeset++

					onceServer(10) {
						if (!event.player.hasDisconnected() && event.player in players) {
							blocks.forValidRefs {
								it.synchronizeToPlayers(false)
							}
						}
					}
				}
			}
		}

		fun onForget(event: ChunkWatchEvent.UnWatch) {
			val subs = playerMap.get(event.level)?.get(event.pos.toLong()) ?: return

			if (subs.players.remove(event.player)) {
				if (subs.isEmpty) {
					playerMap.get(event.level)?.remove(event.pos.toLong())
				} else {
					subs.changeset++
					subs.blockEntities.forValidRefs {
						it.synchronizer.removeEndpointFor(event.player)
					}
				}
			}
		}

		fun playerDisconnected(event: PlayerEvent.PlayerLoggedOutEvent) {
			for (tree in playerMap.values) {
				val iterator = tree.iterator()

				for (entry in iterator) {
					if (entry.value.players.remove(event.entity)) {
						entry.value.changeset++

						if (entry.value.isEmpty) {
							iterator.remove()
						}
					}
				}
			}
		}
	}
}
